#========================================================================================
# package      DE_EPAGES::Object::API::Attributes::LocalizedAttribute
# base         DE_EPAGES::Object::API::Attributes::DefaultAttribute
# state        public
#----------------------------------------------------------------------------------------
# description  Implements localized attributes for objects (types LocalizedString, LocalizedFile).
#========================================================================================
package DE_EPAGES::Object::API::Attributes::LocalizedAttribute;
use base qw( DE_EPAGES::Object::API::BaseAttribute );

use strict;
use DE_EPAGES::Core::API::Log qw (LogDebug);
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger );
use DE_EPAGES::Object::API::Factory qw (LoadObject);
use DE_EPAGES::Object::API::Language qw (GetCurrentLanguageID);
use DE_EPAGES::Object::API::Table::LocalizedStringAttribute qw(
    DeleteLocalizedStringAttribute
    InsertLocalizedStringAttribute
    UpdateLocalizedStringAttribute
    GetInfoLocalizedStringAttributesByObjectID
    GetPKeyLocalizedStringAttributesByObjectID
    GetPKeyLocalizedStringAttributesByAttributeID
);
use DE_EPAGES::Trigger::API::Trigger qw( TriggerHook );

#========================================================================================
# function     _getLocalizedAttributes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hValues = $self->_getLocalizedAttributes( $Object )
#----------------------------------------------------------------------------------------
# description  returns all localized attribute values of an object directly from
#               the database, i.e. without using the object attribute cache.
#----------------------------------------------------------------------------------------
# input        $Object | object | object
# return       $hValues | $hValues->{$AttributeName}->{$LanguageID} = $Value | ref.hash
#========================================================================================
sub _getLocalizedAttributes {
    my $self = shift;
    my ($Object) = @_;

    my %Values;
    foreach ( @{ GetInfoLocalizedStringAttributesByObjectID($Object->id) } ) {
        my $Name = LoadObject($_->{'AttributeID'})->alias;
        $Values{$Name}{$_->{'LanguageID'}} = $_->{'Value'};
    }

    return \%Values;
}

#========================================================================================
# function     getAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = $Package->getAttributes( $Object, $aNames, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Returns multiple attribute values. Don't call this function directly.
#               Use $Object->get() instead.
#----------------------------------------------------------------------------------------
# input        $Object     | object          | object
# input        $aNames     | list of attribute names | ref.array
# input        $LanguageID | language id | integer
# return       $hValues    | attribute name => value hash | ref.hash
#========================================================================================
sub getAttributes {
    my $self = shift;
    my ($Object, $aNames, $LanguageID) = @_;

    return { map { $_ => $self->getAttribute( $Object, $_, $LanguageID) } @$aNames };
}

#========================================================================================
# function     setAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Package->setAttributes( $Object, $hValues, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Sets multiple attribute values. Don't call this function directly.
#               Use $Object->set() instead.
#----------------------------------------------------------------------------------------
# input        $Object     | object          | object
# input        $hValues    | attribute name => value hash | ref.hash
# input        $LanguageID | language id | integer
#========================================================================================
sub setAttributes {
    my $self = shift;
    my ($Object, $hValues, $LanguageID) = @_;

    $self->setAttribute( $Object, $_ => $hValues->{$_}, $LanguageID ) foreach keys %$hValues;
    return;
}

# EPG-23757: non-overloadable version of getAttribute() to be used by setAttribute()
# for sub classes that overload getAttribute()
sub _getAttributeCached {
    my $self = shift;
    my ($Object, $Name, $LanguageID) = @_;

    # return value from cache if exists
    return $Object->{'LocalizedAttributes'}{$Name}{$LanguageID} if exists $Object->{'LocalizedAttributes'};

    # get values from database
    my $hValues = $self->_getLocalizedAttributes( $Object );

    # write values to cache
    $Object->{'LocalizedAttributes'} = $hValues;

    return $hValues->{$Name}->{$LanguageID};
}

#========================================================================================
# function     getAttribute
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Package->getAttribute( $Object, $Name, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Returns a single attribute value. Don't call this function directly.
#               Use $Object->get() instead.
#----------------------------------------------------------------------------------------
# input        $Object     | object          | object
# input        $Name       | attribute name  | string
# input        $LanguageID | language id | int
# return       $Value      | attribute value | string
#========================================================================================
sub getAttribute {
    my $self = shift;
    my ($Object, $Name, $LanguageID) = @_;
    $LanguageID = GetCurrentLanguageID() unless defined $LanguageID;
    TestInteger( 'LanguageID', $LanguageID );

    return $self->_getAttributeCached( $Object, $Name, $LanguageID ) // $self->defaultValue( $Object, $Name, $LanguageID );
}


#========================================================================================
# function     setAttribute
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Package->setAttribute( $Object, $Name, $Value, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Sets a single attribute value. Don't call this function directly.
#               Use $Object->set() instead.
#----------------------------------------------------------------------------------------
# input        $Object     | object          | object
# input        $Name       | attribute name  | string
# input        $Value      | attribute value | string
# input        $LanguageID | language id | int
#========================================================================================
sub setAttribute {
    my $self = shift;
    my ($Object, $Name, $Value, $LanguageID) = @_;
    TestInteger( 'LanguageID', $LanguageID );

    # set value to undef if equal to default value
    if( defined $Value ) {
        my $DefaultValue = $self->defaultValue( $Object, $Name, $LanguageID );
        $Value = undef if defined $DefaultValue && $Value eq $DefaultValue;
    }

    my $Attribute = $Object->class->attribute($Name);
    my $OldValue = $self->_getAttributeCached( $Object, $Name, $LanguageID );
    if (defined $OldValue) {
        if (defined $Value) {
            UpdateLocalizedStringAttribute($Object->id, $Attribute->id, $LanguageID, { Value => $Value });
        }
        else {
            DeleteLocalizedStringAttribute($Object->id, $Attribute->id, $LanguageID);
        }
    }
    elsif (defined $Value) {
        InsertLocalizedStringAttribute({
            ObjectID => $Object->id,
            AttributeID => $Attribute->id,
            LanguageID => $LanguageID,
            Value => $Value,
        });
    }

    # delete the cache
    delete $Object->{'LocalizedAttributes'};

    return;
}

#========================================================================================
# function     defaultValue
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Package->defaultValue( $Object, $Name, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Returns the default value for the attribute.
#               If the stored value is undefined, <function getAttribute>() will return
#               the default value. If the default value is passed to
#               <function setAttribute>, it will delete the stored value.
#----------------------------------------------------------------------------------------
# input        $Object   | object       | object
# input        $Name       | attribute name  | string
# input        $LanguageID | language id | int
#========================================================================================
sub defaultValue { ## no critic qw(RequireCommentForExportedFunctions)
#    my $self = shift;
#    my ( $Object, $Name, $LanguageID ) = @_;
    return undef;
}

#========================================================================================
# function     getAllAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = $Package->getAllAttributes( $Object, $aNames );
#----------------------------------------------------------------------------------------
# description  Returns multiple attribute values. Don't call this function directly.
#               This method is used by DALObject->exportableAttributes().
#               Note: this method returns only the stored values. Default values
#               (see <function defaultValue) are not considered!
#----------------------------------------------------------------------------------------
# input        $Object   | object       | object
# input        $aNames     | list of attribute names (optional) | array ref
# return       $hValues    | attribute name => language => value hash | hash ref
#========================================================================================
sub getAllAttributes {
    my $self = shift;
    my ($Object, $aNames) = @_;

    LogDebug('getAllAttributes', [ $Object, $aNames ]);

    my $hValues = $self->_getLocalizedAttributes($Object);

    TriggerHook('API_GetAllAttributes', {
        'Object' => $Object,
        'Values' => $hValues,
        'Names' => $aNames,
    });

    return $hValues;
}

#========================================================================================
# function     deleteObject
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Package->deleteObject( $Object, $aNames );
#----------------------------------------------------------------------------------------
# description  Deletes named attribute values from the given object.
#----------------------------------------------------------------------------------------
# input        $ObjectID   | object id       | int
# input        $aNames     | attribute names | array ref
#========================================================================================
sub deleteObject {
    my $self = shift;
    my ($Object, $aNames) = @_;

    DeleteLocalizedStringAttribute($_) foreach @{GetPKeyLocalizedStringAttributesByObjectID($Object->id)};
    return;
}

#========================================================================================
# function     deleteAttribute
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Package->deleteAttribute( $Attribute );
#----------------------------------------------------------------------------------------
# description  Deletes an attribute from all objects.
#----------------------------------------------------------------------------------------
# input        $Attribute  | attribute object | object
#========================================================================================
sub deleteAttribute {
    my $self = shift;
    my ($Attribute) = @_;

    DeleteLocalizedStringAttribute($_) foreach @{GetPKeyLocalizedStringAttributesByAttributeID($Attribute->id)};
    return;
}

1;

