#========================================================================================
# §package      DE_EPAGES::Order::API::Order
# §state        public
#----------------------------------------------------------------------------------------
# §description  This package provides the main customer order functions.
#========================================================================================
package DE_EPAGES::Order::API::Order;
use base Exporter;

use strict;
use feature 'switch';
our @EXPORT_OK = qw (
    Basket2Order
    SendOrderConfirmationMail
    SendStatusChangeMail
    SendStatusChangeMails
    GetEditableOrder
    CreateOrder
    CreateOrderWithContainer
    CreateOrderWithoutContainer
    SendOrderDocumentAsPDFMail
    GetLatestOrderInvoice
    GenerateOrderDocumentPDF
    GetLabelImagePath
);

use Config;
use HTML::Entities;

use DE_EPAGES::CodeGenerator::API::TLEProcessor;
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::File qw ( GetFileContent ExistsFile );
use DE_EPAGES::Core::API::Log qw ( SetLogLevel );
use DE_EPAGES::Core::API::PerlTools qw ( GenerateGUID );
use DE_EPAGES::Core::API::TestParameter qw ( TestHash TestObject );
use DE_EPAGES::Core::API::Unicode qw ( Unicode2Filename );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Customer::API::Shop qw ( InsertCustomerForUser );
use DE_EPAGES::EPagesJ::API::PDF::PdfClient;
use DE_EPAGES::MailType::API::TextMail qw ( SendTextMail );
use DE_EPAGES::Object::API::Factory qw ( Transaction );
use DE_EPAGES::Object::API::Language qw ( GetCodeByLanguageID );
use DE_EPAGES::Order::API::Constants qw ( ORDER_STATUS_ATTRIBUTES );
use DE_EPAGES::Order::API::OrderDocument qw ( GetCustomerInformationPDFAttachment GetLegalDocumentPDFAttachment );
use DE_EPAGES::Order::API::Shop qw ( GetNextOrderNo );
use DE_EPAGES::Order::DAL::Basket qw ( DAL_UpdateBasketOrderGUID );
use DE_EPAGES::Presentation::API::PageType qw ( LoadPageTypeByAlias );
use DE_EPAGES::Presentation::API::Template qw ( ProcessPage );
use DE_EPAGES::ShopMailType::API::ShopMailType qw ( GetShopMailType );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Trigger::API::BatchProcessor qw ( RunBatchProcess );

#ERRORDEFINITION
sub ErrorDefinition () {
     return {
        'MISSING_FROM'  => 'No sender e-mail address in shop #ShopAlias'
    }
}
#ERRORDEFINITION END

#========================================================================================
# §function     Basket2Order
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Order = Basket2Order($Basket);
# §syntax       $Order = Basket2Order($Basket, $hOptions);
# §example      $Order = Basket2Order($Basket, { SendConfirmationMail => 1 });
#----------------------------------------------------------------------------------------
# §description  Creates a customer order by a basket and deletes the basket at the end.
#               Addresses will be copied and lineitemcontainer will be moved to new order.
#----------------------------------------------------------------------------------------
# §input        $Basket | basket |  object
# §input        $hOptions | (optional) order creation options.
#                           This hash is passed to the hooks triggerd by this function.
#               <ul>
#                   <li>SendConfirmationMail - trigger mail event 'CustomerOrderConfirmation',
#                                               which sends order confirmation e-mail
#                                               (optional, default=1) - boolean</li>
#               </ul> | ref.hash
# §return       $Order | new customer order | object
# §hook         API_Basket2Order | triggered in basket to order transaction, hook parameter keys :
#               <ul>
#                 <li>Basket  - basket - object</li>
#                 <li>Order   - customer order - object</li>
#                 <li>Options - order creation options - ref.hash</li>
#               </ul>
# §hook         API_OrderCreated | triggered after order creation, hook parameter keys :
#               <ul>
#                 <li>Order   - customer order - object</li>
#                 <li>Options - order creation options - ref.hash</li>
#               </ul>
#========================================================================================
sub Basket2Order ($;$) {
    my ($Basket, $hOptions) = @_;

    my $Order;
    RunBatchProcess(['CreateOrders', 'UpdateProducts'], sub {

        $hOptions = {} unless defined $hOptions;
        $hOptions->{'SendConfirmationMail'} = 1 unless defined $hOptions->{'SendConfirmationMail'};

        my $User      = $Basket->get('User');
        my $Shop     = $Basket->getSite();
        my $hBasketRollback = $Basket->get([qw(Parent Alias OrderGUID)]);

        #--- check if billing address exists
        my $BillingAddress = $Basket->get('BillingAddress');
        Error('NoBillingAddress', {'Basket' => $Basket->get('Path')}) unless defined $BillingAddress;

        # block basket
        my $OrderGUID = GenerateGUID();
        DAL_UpdateBasketOrderGUID($Basket->id, $OrderGUID); # set database entry direct to avoid race conditions


        # sleep(10);  # uncomment this line for testing parallel running order creation processes (EPG-25327)

        # set object attribute and reset cache for object on other AS
        # OrderGUID was already set in the database by DAL_UpdateBasketOrderGUID,
        # but it's not yet in the cache
        eval {
            Transaction( sub {
                $Basket->set({
                    'Parent'    => $User->folder('OrderedBaskets'),
                    'Alias'     => $OrderGUID,
                    'OrderGUID' => $OrderGUID
                });
            });

            # get customer, if not exists create customer
            my $Customer = $Basket->get('Customer');
            Transaction( sub {
                if (not defined $Customer) {
                    my $folder = $User->folder('Addresses');
                    my $_BillingAddress = $Basket->get('BillingAddress')->clone({
                        'Alias' => $folder->nextAlias('BillingAddress %s'),
                        'Parent'=> $folder,
                    });
                    $User->set({'BillingAddress'=>$_BillingAddress});

                    if ($Basket->get('ShippingAddress')) {
                        my $ShippingAddress = $Basket->get('ShippingAddress')->clone({
                            'Alias' => $folder->nextAlias('Address %s'),
                            'Parent'=> $folder,
                        });
                        $User->set({'ShippingAddress'=>$ShippingAddress});
                    }
                    $Customer = InsertCustomerForUser($User, $Shop);
                    my $Container = $Basket->get('LineItemContainer');
                    $Customer->set({'TaxModel' => $Container->get('TaxModel')}) if defined $Container && $Customer->get('TaxModel') != $Container->get('TaxModel');
                }

                $Customer->folder('Orders'); # make sure the Orders folder exists before the next transaction
            });

            my $OrderNo = GetNextOrderNo($Shop, $Customer);

            # create order
            Transaction( sub {
                # create order, if the next order number is already used, try it again
                my $tryit = 1;
                my $trycount = 5;
                while ($tryit) {
                    $tryit = 0;
                    $trycount--;
                    eval {
                        $Order = CreateOrderWithoutContainer( $Customer, {
                            'Alias' => $OrderNo,
                            'User' => $User,
                            'GUID' => $OrderGUID,
                        });
                    };
                    if (ExistsError()) {
                        my $Error = GetError();
                        Error($Error) unless $Error->code eq 'DBIExecute' and $Error->vars->{'code'} == 2601 and $trycount; ## no critic qw(ProhibitMagicNumbers) # double key
                        $OrderNo = GetNextOrderNo($Shop, $Customer);
                        $tryit = 1;
                    }
                }

                # move children of basket to order second step (move objects to order)
                my $LineItemContainer = $Basket->get('LineItemContainer');
                $LineItemContainer->set({'Parent' => $Order});

                my %OrderAttributes = (
                    'CustomerComment' => $Basket->get('CustomerComment'),
                    'LineItemContainer' => $LineItemContainer ,
                );

                foreach my $Attribute (qw( BillingAddress ShippingAddress ShopAddress)) {
                    my $Value = $Basket->get($Attribute);
                    next unless defined $Value;
                    $OrderAttributes{$Attribute} = $Value->clone({ 'Parent'=>$Order, 'Alias'=> $Attribute });
                }
                $Order->set(\%OrderAttributes);

                # activate the user account (EPG-11705), not for Unity shops (UNITY-5001)
                $User->set({
                    'IsActivated' => 1,
                    'IsEmailVerified' => 1,
                }) unless ($User->get('IsActivated') || $User->get('IsAnonymous') || $Shop->isUnity());

                # Trigger hook, that can abort the order creation (via Error)
                # e.g. for synchronous data exchange with third party systems
                TriggerHook('API_Basket2Order', {
                    'Order'   => $Order,
                    'Basket'  => $Basket,
                    'Options' => $hOptions
                });

                # dont remove basket, move to ordered folder
                $Basket->set({
                    'LineItemContainer' => undef,
                });
            });

            $Customer->set({'TaxArea' => $Order->container->get('TaxArea')}) if $Shop->get('CanSelectTaxAreaOnBasket');
        };
        if (ExistsError()){
            my $Error = GetError();
            $Basket->set($hBasketRollback);
            Error($Error);
        }

        $hOptions->{'callerComponent'} = 'AS';
        TriggerHook('API_OrderCreated', {
            'Order'   => $Order,
            'Options' => $hOptions
        });
    });

    return $Order;
}

#========================================================================================
# §function     SendOrderConfirmationMail
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       SendOrderConfirmationMail($Order);
#----------------------------------------------------------------------------------------
# §description  Send a confirmation mail to customer (Order.BillingAddress.EMail is used).
#               The activated shop mail type 'CustomerOrderConfirmation' is used to define
#               the content of the mail.
#----------------------------------------------------------------------------------------
# §input        $Order | customer order | object
#========================================================================================
sub SendOrderConfirmationMail ($) {
    my ($Order) = @_;

    #send order confirmation mail to customer
    my $Shop = $Order->getSite();
    my $MailType = GetShopMailType( $Shop, 'CustomerOrderConfirmation' );

    # don't event try to send an e-mail if the merchant doesn't have a sender e-mail
    my $From = $MailType->from;

    unless( defined $From && $From =~ /\S/ ) {
        Warning('MISSING_FROM', {'ShopAlias' => $Shop->alias});
        return;
    }

    my $BillingAddress = $Order->get('BillingAddress') ? $Order->get('BillingAddress'): $Order->get('Customer')->get('BillingAddress');
    my $Email = $BillingAddress->get('EMail');

    # create and send mail
    my %MailParams = (
        'HTMLAllowed' => defined $Order->get('User') ? $Order->get('User')->get('IsHtmlEMailAllowed') : 1,
        'Attachments' => [],
    );
    $MailParams{'TO'} = $Email if defined $Email;

    $MailParams{'Attachments'} = _getLegalAttachements($Order, $MailType);

    my %Vars = (
        IsUnityEmailConfirmation => $Shop->get('IsUnityRegistered'),
    );
    my $Container = $Order->get('LineItemContainer');
    my %RegionalParams = (
        Locale      => $Container->get('LocaleID'),
        Language    => GetCodeByLanguageID($Container->get('LanguageID')),
    );

    # allow lineitems to manipulate mail
    my $LineItems = $Container->get('Positions');
    $_->addMailParams(\%MailParams, \%RegionalParams) foreach @$LineItems;
    $MailType->send( $Order, \%MailParams, \%RegionalParams, \%Vars );
    $_->clearMail(\%MailParams, \%RegionalParams) foreach @$LineItems;
    return;
}

#========================================================================================
# §function     SendStatusChangeMails
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       SendStatusChangeMails($Order, $hOldStatus, $hNewStatus);
#----------------------------------------------------------------------------------------
# §description  Sends order status notification e-mails to the customer if a status was
#               set (from null to not null).
#               The e-mail address Order.BillingAddress.EMail is used.
#               See also <function SendStatusChangeMail>.
#----------------------------------------------------------------------------------------
# §input        $Order | customer order | object
# §input        $hOldStatus | pevious status change dates | ref.hash.
# §input        $hStatus | new status change dates (name => DateTime).
#                   This hash may contain other attribute values, but only the status
#                   attributes defined in <package DE_EPAGES::Order::API::Constants>
#                   will be used.
#                   | ref.hash
#========================================================================================
sub SendStatusChangeMails {
    my ($Order, $hOldStatus, $hNewStatus) = @_;
    TestObject('Order', $Order);
    TestHash('hOldStatus', $hOldStatus);
    TestHash('hNewStatus', $hNewStatus);

    # send emails if the status has changed
    foreach my $Status ( ORDER_STATUS_ATTRIBUTES ) {
        next unless defined $hNewStatus->{$Status};
        next if defined $hOldStatus->{$Status};
        SendStatusChangeMail($Order, $Status);
    }
}

#========================================================================================
# §function     SendStatusChangeMail
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       SendStatusChangeMail($Order, $Event);
#----------------------------------------------------------------------------------------
# §description  Send a status changed mail to customer (Order.BillingAddress.EMail is used).
#               The activated shop mail type 'CustomerOrderStatus$Event' is used to define
#               the content of the mail. This mail will not be send on event 'ArchivedOn'.
#----------------------------------------------------------------------------------------
# §input        $Order | customer order | object
# §input        $Event | name of order status | string
#========================================================================================
sub SendStatusChangeMail ($$) {
    my ($Order, $Event) = @_;

    return if $Event eq 'ArchivedOn';

    #send order confirmation mail to customer
    my $MailType = GetShopMailType( $Order->getSite(), 'CustomerOrderStatus'.$Event, 1 );

    return unless defined $MailType;
    return unless $Order->get('Customer');

    my $BillingAddress = $Order->get('BillingAddress') ? $Order->get('BillingAddress'): $Order->get('Customer')->get('BillingAddress');

    # EPG-21431 prohibit that email is send without an recipient
    return unless $BillingAddress->get('EMail');

    # create and send mail
    my %MailParams = (
        'TO'          => $BillingAddress->get('EMail'),
        'HTMLAllowed' => $Order->get('Customer')->get('IsHtmlEMailAllowed'),
        'SkipBadRecipients' => 0,
    );


    my $ahAttachments = [];

    my $hPdfContent;
    if ($MailType->get('OrderDocumentPdfAttachment')) {
        my $LatestOrderInvoice = GetLatestOrderInvoice($Order);
        if(defined $LatestOrderInvoice){
            $hPdfContent = GenerateOrderDocumentPDF($LatestOrderInvoice);
                push @$ahAttachments, {
                    'filename' => $hPdfContent->{'filename'},
                    'data'     => $hPdfContent->{'pdf'},
                    'encoding' => 'base64'
                };
        }
    }

    push @$ahAttachments, @{ _getLegalAttachements($Order, $MailType) };

    $MailParams{'Attachments'} = $ahAttachments;

    my %Vars = ();
    my $Container = $Order->get('LineItemContainer');
    Error('NoLineItemConatiner') unless defined $Container;
    my %RegionalParams = (
        Locale   => $Container->get('LocaleID'),
        Language => GetCodeByLanguageID($Container->get('LanguageID')),
    );
    $MailType->send( $Order, \%MailParams, \%RegionalParams, \%Vars );

    return;
}

sub _getLegalAttachements {
    my ($Order, $MailType) = @_;

    my @Attachments;
    my $Shop = $Order->getSite();
    my $LanguageID = $Order->get('LineItemContainer')->get('LanguageID');

    my $LegalDocs = $Shop->isUnity()
        ? $MailType->get('UnityAttachLegalDocumentsPDF')
        : $MailType->get('AttachCustomerInformationPDF')
            ? [$Shop->get('CustomerInformation')]
            : undef;

    foreach my $LegalDocument (@$LegalDocs) {
        my $hLegalDocumentPDFAttachment = GetLegalDocumentPDFAttachment($Shop, $LanguageID, $LegalDocument);
        push @Attachments, $hLegalDocumentPDFAttachment if $hLegalDocumentPDFAttachment;
    }

    return \@Attachments;
}

#========================================================================================
# §function     GetEditableOrder
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $EditableOrder = GetEditableOrder($Order);
#----------------------------------------------------------------------------------------
# §description  Returns the editable order (copy and possible changed copy of order). If
#               no editable order exists this function returns undef.
#----------------------------------------------------------------------------------------
# §input        $Order | customer order | object
# §return       $EditableOrder | copy of order (optional) | string
#========================================================================================
sub GetEditableOrder ($) {
    my ($Order) = @_;

    my $Customer = $Order->get('Customer');
    return undef unless $Customer->existsChild('EditableOrders');
    my $EditableOrders = $Customer->child('EditableOrders');
    my $EditableOrderID = $EditableOrders->findChild($Order->alias);
    return undef unless defined $EditableOrderID;
    return $EditableOrders->child($Order->alias);
}

#========================================================================================
# §function     CreateOrder
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Order = CreateOrder($Customer, $hVars);
#----------------------------------------------------------------------------------------
# §description  Creates an empty order which will belong to the given customer.
#               Addresses are copied and an empty lineitemcontainer is added.
#               Sets the default shipping and payment method.
#----------------------------------------------------------------------------------------
# §input        $Customer | customer |  object
# §input        $hVars | <ul>
#                   <li>CurrencyID - (optional; default: shop default currency) alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel - (optional; default: customer tax model) gross (1) or net (0) price  - int
#               </ul> | ref.hash (optional)
# §return       $Order | new customer order | object
#========================================================================================
sub CreateOrder ($;$) {
    my ($Customer, $hVars) = @_;
    TestObject( 'Customer', $Customer, 'DE_EPAGES::Customer::API::Object::Customer' );

    $hVars //= {};
    my $Shop = $Customer->getSite();

    my $CurrencyID = $hVars->{'CurrencyID'} // $Shop->get('CurrencyID');
    my $TaxModel = $hVars->{'TaxModel'} // $Customer->get('TaxModel');

    # get customer addresses
    my $BillingAddress = $Customer->get('BillingAddress');
    Error('NoBillingAddress', {'Customer' => $Customer->get('Path')}) unless defined $BillingAddress;

    # get default shipping method
    my @ShippingMethods = grep { $_->get('IsActivated') and $_->get('ShippingType')->get('IsActivated') }  @{$Shop->get('ShippingMethods')};
    Error('NoShipping') unless scalar @ShippingMethods;

    my ($DefaultShippingMethod) = grep { $_->get('IsDefault') } @ShippingMethods;
    $DefaultShippingMethod //= $ShippingMethods[0]; # select first one if no method is marked as default

    if (!$DefaultShippingMethod->usableAtStorefront({'CurrencyID' => $CurrencyID, 'TaxModel' => $TaxModel})) {
        # if not usable at storefront, find another usable one
        foreach my $ShippingMethod (@ShippingMethods) {
            if ($ShippingMethod->usableAtStorefront({'CurrencyID' => $CurrencyID, 'TaxModel' => $TaxModel})) {
                $DefaultShippingMethod = $ShippingMethod;
                last;
            }
        }
    }

    # get default payment method

    my @PaymentMethods = grep {
        $_->get('IsActivated')
        and $_->get('PaymentType')->get('IsActivated')
        and not($_->get('PaymentType')->get('IsCustomerRequired')) # in MBO only not customer requiring payment methods could be choosen
    } @{$DefaultShippingMethod->get('PaymentMethods')};
    Error('NoPayment') unless scalar @PaymentMethods;

    my ($DefaultPaymentMethod) = grep { $_->get('IsDefault') } @PaymentMethods;
    $DefaultPaymentMethod //= $PaymentMethods[0];  # select first one if no method is marked as default
    if (!$DefaultPaymentMethod->usableAtStorefront({'CurrencyID' => $CurrencyID, 'TaxModel' => $TaxModel})) {
        # if not usable at storefront, find another usable one
        foreach my $PaymentMethod (@PaymentMethods) {
            if ($PaymentMethod->usableAtStorefront({'CurrencyID' => $CurrencyID, 'TaxModel' => $TaxModel})) {
                $DefaultPaymentMethod = $PaymentMethod;
                last;
            }
        }
    }

    my %CopyVars = (
        %$hVars,
        'Alias' => GetNextOrderNo($Shop, $Customer), # generate next order number before the transaction to avoid long locks
    );
    my $Order;
    Transaction( sub {

        $Order = CreateOrderWithContainer( $Customer, \%CopyVars );

        my $CloneAddress = $BillingAddress->clone({ 'Parent' => $Order, 'Alias' => 'BillingAddress' });
        $Order->set({ 'BillingAddress' => $CloneAddress });

        if($Shop->get('TaxMethod')){
            $Shop->get('TaxMethod')->updateContainerTaxArea($Order->container);
        }else {
            $Order->updateTaxArea;
            $Order->updateTaxAreaDigital;
        }
        # insert payment if not already exists
        my $Container = $Order->get('LineItemContainer');
        $DefaultPaymentMethod->addToContainer($Container, {}) unless $Container->get('Payment');
    });
    return $Order;
}

#========================================================================================
# §function     CreateOrderWithContainer
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Order = CreateOrderWithContainer( $Customer, $hValues );
# §example      Transaction( sub {
#                   $Order = CreateOrderWithContainer( $Customer, { 'Alias' => '1001' } );
#                   $Order->set({
#                       'BillingAddress' => $Customer->get('BillingAddress')->clone({ 'Parent' => $Order, 'Alias' => 'BillingAddress' }),
#                   });
#               });
#----------------------------------------------------------------------------------------
# §description  Creates an empty order for a customer without shipping and payment
#               methods and without billing/shipping address.
#               Creates a lineitem container and sets the shop address.
#----------------------------------------------------------------------------------------
# §input        $Customer | customer |  object
# §input        $hValues | <ul>
#                   <li>CurrencyID - (optional; default: user currency or shop default currency) alpha currency code (ISO 4217) - char(3)
#                   <li>LocaleID - (optional; default: user locale or shop default locale) locale id - string
#                   <li>TaxArea - (optional; default: customer tax area) tax area - DE_EPAGES::Tax::API::Object::TaxArea
#                   <li>TaxModel - (optional; default: customer tax model) gross (1) or net (0) price  - int
#                   <li>User - (optional; default: first user of customer or undef) user - object
#                   <li>Alias - (optional; default: generate new order number) order number - string
#               </ul> | ref.hash (optional)
# §return       $Order | new customer order | object
#========================================================================================
sub CreateOrderWithContainer {
    my ($Customer, $hValues) = @_;

    $hValues //= {};
    my $Shop = $Customer->getSite();

    my $CurrencyID = $hValues->{'CurrencyID'} // $Shop->get('CurrencyID');
    my $TaxArea = $hValues->{'TaxArea'} // $Customer->get('TaxArea');
    my $TaxModel = $hValues->{'TaxModel'} // $Customer->get('TaxModel');

    my $NextOrderNo = $hValues->{'Alias'} // GetNextOrderNo($Shop, $Customer);

    my $Order;
    Transaction( sub {
        # create a new order
        $Order = CreateOrderWithoutContainer( $Customer, {
            'Alias' => $NextOrderNo,
            'User' => $hValues->{'User'},
        });

        my $User = $Order->get('User');
        my $LocaleID = $hValues->{'LocaleID'} // (defined $User ? $User->get('LocaleID') : $Shop->get('LocaleID'));

        # add  line item container
        $Order->addLineItemContainer({
            'LocaleID'      => $LocaleID,
            'CurrencyID'    => $CurrencyID,
            'TaxArea'       => $TaxArea,
            'TaxModel'      => $TaxModel,
        });

        # set the shop address
        my $ShopAddress = $Shop->get('Address');
        my $CloneShopAddress = defined $ShopAddress ? $ShopAddress->clone({ 'Parent'=>$Order, 'Alias'=> 'ShopAddress' }) : undef;
        $Order->set({ 'ShopAddress' => $CloneShopAddress  });
    });
    return $Order;
}

#========================================================================================
# §function     CreateOrderWithoutContainer
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Order = CreateOrderWithoutContainer( $Customer, $hValues );
# §example      Transaction( sub {
#                   $Order = CreateOrderWithoutContainer( $Customer, { 'Alias' => '1001' } );
#                   $Order->addLineItemContainer;
#                   $Order->set({
#                       'BillingAddress' => $Customer->get('BillingAddress')->clone({ 'Parent' => $Order, 'Alias' => 'BillingAddress' }),
#                       'ShopAddress' => $Shop->get('Address')->clone({ 'Parent'=>$Order, 'Alias'=> 'ShopAddress' }),
#                   });
#               });
#----------------------------------------------------------------------------------------
# §description  Creates an empty order which for a customer without line item container
#               and without shop address.
#----------------------------------------------------------------------------------------
# §input        $Customer | customer |  object
# §input        $hValues | <ul>
#                   <li>User - (optional; default: first user of customer or undef) user - object
#                   <li>Alias - (optional; default: generate new order number) order number - string
#                   <li>GUID - (optional; default: generate new GUID) GUID - string
#               </ul> | ref.hash (optional)
# §return       $Order | new customer order | object
#========================================================================================
sub CreateOrderWithoutContainer {
    my ($Customer, $hValues) = @_;

    $hValues //= {};
    my $Shop = $Customer->getSite();

    my $User = $hValues->{'User'} // $Customer->get('Users')->[0];

    my $NextOrderNo = $hValues->{'Alias'} // GetNextOrderNo($Shop, $Customer);
    my $Order;
    Transaction( sub {
        # create a new order
        my %Values = (
            'Alias'         => $NextOrderNo,
            'Shop'          => $Shop,
            'Customer'      => $Customer,
            'User'          => $User,
        );
        $Values{'GUID'} = $hValues->{'GUID'} if defined $hValues->{'GUID'};

        $Order = $Customer->folder('Orders')->insertChild('CustomerOrder', \%Values );
    });
    return $Order;
}

#========================================================================================
# §function     SendOrderDocumentAsPDFMail
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       SendOrderDocumentAsPDFMail($OrderDocument, $Servlet);
#----------------------------------------------------------------------------------------
# §description  Send an order document as pdf file mail to customer (Order.BillingAddress.EMail
#               is used). The activated shop mail type 'OrderDocumentPDF' is used to define the
#               content of the mail.
#               The OrderDocument must be Invoice, PackingSlip, CreditNote, or ShippingLabel.
#----------------------------------------------------------------------------------------
# §input        $OrderDocument | order document | object
# §input        $Servlet | the servlet | object
#========================================================================================
sub SendOrderDocumentAsPDFMail($$){
    my ($OrderDocument, $Servlet) = @_;

    my $Order = $OrderDocument->get('Order');
    my $Shop = $Order->getSite;
    my $From  = $Shop->get('MailFrom');

    unless( defined $From && $From =~ /\S/ ) {
        Warning('MISSING_FROM', {'ShopAlias' => $Shop->alias});
        return;
    }

    my $BillingAddress = $Order->get('BillingAddress');
    my $DefaultMail = $BillingAddress ? $BillingAddress->get('EMail') : undef;

    my $EmailTo  = $Servlet->params('To') || $DefaultMail;
    my $EmailCc  = $Servlet->params('Cc');
    my $EmailBcc = $Servlet->params('Bcc');
    my $Subject  = $Servlet->params('Subject');
    my $EmailMsg = $Servlet->params('Msg');

    my $hPdfContent = GenerateOrderDocumentPDF($OrderDocument);

    my $aAttachments = [
        {
            'filename' => $hPdfContent->{'filename'},
            'data'     => $hPdfContent->{'pdf'},
            'encoding' => 'base64'
        }
    ];

    if (!_IsShippingLabel($OrderDocument)) {
        my $OrderDocumentType = $OrderDocument->class->alias;
        if ($Shop->get("OrderDocumentsAttachCustomerInformationPDFTo${OrderDocumentType}Email")) {
            my $hCustomerInformationPDFAttachment = GetCustomerInformationPDFAttachment($Order->getSite, $Order->get('LineItemContainer')->get('LanguageID'));
            push @$aAttachments, $hCustomerInformationPDFAttachment if $hCustomerInformationPDFAttachment;
        }
    }

    # create and send mail
    my %MailParams = (
        'from'        => $From,
        'to'          => $EmailTo,
        'cc'          => $EmailCc,
        'bcc'         => $EmailBcc,
        'subject'     => $Subject,
        'content'     => $EmailMsg,
        'attachments' => $aAttachments,
    );

    SendTextMail(\%MailParams);
    return;
}

#========================================================================================
# §function     GenerateOrderDocumentPDF
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $hPdfContent = GenerateOrderDocumentPDF($OrderDocument);
#----------------------------------------------------------------------------------------
# §description  Converts order document (invoice, packing slip, credit note or shipping
#               label) to pdf document.
#----------------------------------------------------------------------------------------
# §input        $OrderDocument | OrderDocument | object
#----------------------------------------------------------------------------------------
# §return       $hPdfContent | <ul>
#                               <li>pdf       - pdf document                  - binary
#                               <li>filename  - name of pdf file              - string
#                              </ul> | ref.hash
#========================================================================================
sub GenerateOrderDocumentPDF ($) {
    my ($OrderDocument) = @_;

    TestObject('OrderDocument', $OrderDocument);

    my $PdfFilePath = GetLabelImagePath($OrderDocument);
    my $PdfBinary;

    if (_IsShippingLabel($OrderDocument) && ExistsFile($PdfFilePath)) {

        $PdfBinary = ${GetFileContent($PdfFilePath, {'binary'=>1})};

    }
    else {

        my $Shop = $OrderDocument->getSite();
        my $Container = $OrderDocument->get('LineItemContainer');
        my $LanguageID = $Container->get('LanguageID');
        my $LanguageCode = GetCodeByLanguageID($LanguageID);

        my $hVars = $OrderDocument->tleHash;
        $hVars->{'INPUT'} = { 'Locale' => $Container->get('LocaleID') };
        $hVars->{'PDF'} = 1;
        $hVars->{'OSNAME'} = $Config{'osname'};

        my $PageType;
        if ($Shop->get('IsUnityRegistered')) {
            $hVars->{'IsUnityInvoiceDocumentPDF'} = 1;
            $PageType = LoadPageTypeByAlias('UnityMBO-Print' . $OrderDocument->class->alias);
        } else {
            $PageType = LoadPageTypeByAlias('MBO-Print' . $OrderDocument->class->alias);
        }
        my $ObjectPageType = DE_EPAGES::Presentation::API::Object::ObjectPageType->load($PageType, $OrderDocument);
        my $HtmlContent;
        SetLogLevel('WARN', sub {
            $HtmlContent = ProcessPage($OrderDocument, $ObjectPageType, $LanguageCode, $hVars);
        }, 'DE_EPAGES::Presentation::API::Template::INCLUDE');

        my $PdfClient = DE_EPAGES::EPagesJ::API::PDF::PdfClient->create($OrderDocument->getSite);
        $PdfBinary = $PdfClient->transform($HtmlContent);

    }

    return {
      pdf => $PdfBinary,
      filename => _GetOrderDocumentPdfFileName($OrderDocument)
    };

}

#========================================================================================
# §function     _IsShippingLabel
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _IsShippingLabel($OrderDocument);
#----------------------------------------------------------------------------------------
# §description  Returns true if the OrderDocument is a shipping label; false otherwise
#               (invoice, packing slip, credit note, ...)
#----------------------------------------------------------------------------------------
# §input        $OrderDocument | OrderDocument | object
#----------------------------------------------------------------------------------------
# §return       $IsShippingLabel | Returns if the OrderDocument is a Shipping label | boolean
#========================================================================================
sub _IsShippingLabel($){
    my ($OrderDocument) = @_;

    return $OrderDocument->instanceOf('ShippingLabel');

}

#========================================================================================
# §function     GetLabelImagePath
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Path = GetLabelImagePath($ShippingLabel, $FileExtension);
# §example      my $Path  = GetLabelImagePath($ShippingLabel, 'pdf');
#----------------------------------------------------------------------------------------
# §description  Sets and returns the path for the barcode image of a shipping label.
#               Used by USPS when issueing (triggering) a shipping label (to store the
#               recieved pdf) and when sending it by email (from the MBO order documents
#               page). It could also be used by SEUR (since it has its own but exaclty
#               the same function) and other delivery cartridges...
#----------------------------------------------------------------------------------------
# §input        $ShippingLabel | shipping label object | object
# §input        $FileExtension | image file extension | string
# §return       $LabelImagePath | path and image name | string
#========================================================================================
sub GetLabelImagePath {
    my ($ShippingLabel, $FileExtension) = @_;

    my $LabelImagePath = sprintf '%s/barcode_%s.%s',
                                 $ShippingLabel->get('PrivatePath'),
                                 $ShippingLabel->id(),
                                 $FileExtension // 'pdf';

    return $LabelImagePath;
}

#========================================================================================
# §function     _GetOrderDocumentPdfFileName
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _GetOrderDocumentPdfFileName($OrderDocument);
#----------------------------------------------------------------------------------------
# §description  Returns the file name for the pdf order document
#----------------------------------------------------------------------------------------
# §input        $OrderDocument | OrderDocument | object
#----------------------------------------------------------------------------------------
# §return       $PdfFileName | PdfFileName | string
#========================================================================================
sub _GetOrderDocumentPdfFileName($){
    my ($OrderDocument) = @_;

    my $Order = $OrderDocument->get('Order');
    my $LanguageID = $Order->get('LineItemContainer')->get('LanguageID');
    my $DocType = $OrderDocument->class()->get('NameOrAlias', $LanguageID);
    my $DocName = $OrderDocument->alias();
    my $FileName = Unicode2Filename($DocType.'_'.$DocName);
    return $FileName.'.pdf';
}

#========================================================================================
# §function     GetLatestOrderInvoice
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       GetLatestOrderInvoice($Order);
#----------------------------------------------------------------------------------------
# §description  Returns the latest invoice document of an order
#----------------------------------------------------------------------------------------
# §input        $Order | Order | object
#----------------------------------------------------------------------------------------
# §return       $Invoice | invoice | object
#========================================================================================
sub GetLatestOrderInvoice{
    my ($Order) = @_;

    # Get latest order document of type invoice from order's folder 'documents'
    my @Invoices = grep {$_->class()->alias() eq 'Invoice'} @{$Order->get('Documents')};

    my $LatestInvoice = undef;
    if(scalar @Invoices){
        my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new();
        @Invoices = sort { $Formatter->format_datetime($b->get('CreationDate'), 'datetime', '%Y%m%d%H%M%S')
                             <=>
                           $Formatter->format_datetime($a->get('CreationDate'), 'datetime', '%Y%m%d%H%M%S')
                         } @Invoices;
        $LatestInvoice = $Invoices[0];
    }
    return $LatestInvoice;
}

1;
