#========================================================================================
# package      DE_EPAGES::Shop::API::TLE::SystemHandler
# state        public
#----------------------------------------------------------------------------------------
# description  Implements system-wide TLE variables for system
#               <ul>
#                   <li>VideoIDs</li>
#               </ul>
#========================================================================================
package DE_EPAGES::Shop::API::TLE::SystemHandler;

use strict;

use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::TestParameter qw ( TestString );
use DE_EPAGES::Object::API::Language qw ( GetCurrentLanguageID );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Shop::API::Constants qw ( TUTORIAL_VIDEO_ATTRIBUTES $UNITY_RECOMMENDED_IMAGE_SIZES );

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'SystemHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Object::API::TLE::SystemHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Object::API::TLE::SystemHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'SystemHandler.register' );

    $Processor->registerHandler('VariableHandler', $self, 'TutorialVideoIDs');
    $Processor->registerHandler('FunctionHandler', $self, 'UnityRecommendedImageSizes');
    return;
}

#========================================================================================
# function     tle
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->tle();
# example      my $Value = $Handler->tle();
#----------------------------------------------------------------------------------------
# description  This function is called by tle processor to get the value of a tle
#               variable.
#----------------------------------------------------------------------------------------
# return       $aVideoIDs | list of video IDs | ref.array
#========================================================================================
sub tle {
    my $self = shift;

    my $UnityVideoTutorialFolder = LoadRootObject()->get('UnityVideoTutorialFolder');

    my $hInfo = $UnityVideoTutorialFolder->get(TUTORIAL_VIDEO_ATTRIBUTES(), GetCurrentLanguageID());

    my @VideoIDs;
    foreach my $VideoAttribute (@{TUTORIAL_VIDEO_ATTRIBUTES()}) {
        if (defined $hInfo->{$VideoAttribute}) {
            push @VideoIDs, $hInfo->{$VideoAttribute};
        }
    }

    return \@VideoIDs;
}

#========================================================================================
# function     UnityRecommendedImageSizes
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnityRecommendedImageSizes(#ImageType)
# example      #LOCAL("ImageSizeText", #UnityRecommendedImageSizes('Product_Image'))
#                 #ImageSizeText
#               #ENDLOCAL
#----------------------------------------------------------------------------------------
# description  Returns the recommended minimal image size for the given image type.
#----------------------------------------------------------------------------------------
# input        #ImageType | type of image  (see DE_EPAGES::Shop::API::Constants::UNITY_RECOMMENDED_IMAGE_SIZES) | integer
# return       #ImageSizeText | text for minimal image size recommendation  | string
#========================================================================================
sub UnityRecommendedImageSizes {
    my $self = shift;
    my ($Processor, $raParams) = @_;
    my ($ImageType) = @$raParams;

    TestString( 'ImageType', $ImageType );

    my $ImageSizeText = exists $UNITY_RECOMMENDED_IMAGE_SIZES->{$ImageType} ? $UNITY_RECOMMENDED_IMAGE_SIZES->{$ImageType} : undef; 
    TestString( 'ImageSizeText', $ImageSizeText );

    return $ImageSizeText;
}

1;