#========================================================================================
# package      DE_EPAGES::Spreedly::API::Object::LineItemPaymentSpreedly
# state        public
# base         DE_EPAGES::ExternalPayment::API::Object::LineItemPaymentExternal
#----------------------------------------------------------------------------------------
# description  object interface for LineItemPaymentSpreedly.
#========================================================================================
package DE_EPAGES::Spreedly::API::Object::LineItemPaymentSpreedly;
use base DE_EPAGES::ExternalPayment::API::Object::LineItemPaymentExternal;

use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime isValid );
use DE_EPAGES::Spreedly::UI::BasketOrder;
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Spreedly::API::Constants qw ( CREDIT_CARD PAYMENT_STATUS_PAID );

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $Package->new('ObjectID' => $ObjectID, ... );
#----------------------------------------------------------------------------------------
# description  overloads DE_EPAGES::Core::API::BaseObject->new
#               to decrypt attributes before store in cache
#----------------------------------------------------------------------------------------
# input        %Attributes | attribute value hash | hash
# return       $Object | existing object | object
#========================================================================================
sub new {
    my $self = shift;
    my (%options) = @_;

    $self->decryptBaseAttributes(\%options);

    return $self->SUPER::new(%options);
}


#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $LineItemPaymentSpreedly->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className {
    return 'LineItemPaymentSpreedly';
}

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $LineItemPaymentSpreedly->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSpreedly = DE_EPAGES::Spreedly::API::Object::LineItemPaymentSpreedly->insert($hInfo);
# example      my $Class = LoadClassByAlias('LineItemPaymentSpreedly');
#               my $LineItemPaymentSpreedly = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertLineItemPaymentSpreedly' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $LineItemPaymentSpreedly | new object | object
# hook         OBJ_InsertLineItemPaymentSpreedly | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSpreedlyID - line item payment id - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertLineItemPaymentSpreedly', {
        'Object' => $self,
        'LineItemPaymentSpreedlyID' => $self->id(),
    });
    delete $hInfo->{$_} foreach qw ( PaymentInformationToken TransactionToken );

    return $self;
}



#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSpreedly->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteLineItemPaymentSpreedly' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteLineItemPaymentSpreedly | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSpreedlyID - line item payment id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteLineItemPaymentSpreedly', {
        'Object' => $self,
        'LineItemPaymentSpreedlyID' => $self->id(),
    });

    $self->SUPER::delete;
    return;
}



#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSpreedly->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateLineItemPaymentSpreedly'
#               and 'OBJ_AfterUpdateLineItemPaymentSpreedly'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateLineItemPaymentSpreedly | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSpreedlyID - line item payment id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateLineItemPaymentSpreedly | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSpreedlyID - line item payment id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateLineItemPaymentSpreedly', {
        'Object'=>$self,
        'Values'=> $hValues,
        'LineItemPaymentSpreedlyID' => $self->id(),
        'LanguageID' => $LanguageID,
    });

    $self->SUPER::set($hValues, $LanguageID);

    TriggerHook('OBJ_AfterUpdateLineItemPaymentSpreedly', {
        'Object'=>$self,
        'Values'=> $hValues,
        'LineItemPaymentSpreedlyID' => $self->id(),
        'LanguageID' => $LanguageID,
    });

    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Spreedly::API::Object::LineItemPaymentSpreedly->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
    };
}

#========================================================================================
# function     _updateTransactionStatus
# state        private
#----------------------------------------------------------------------------------------
# syntax       _updateTransactionStatus($Status);
#----------------------------------------------------------------------------------------
# description  Updates LineItemPaymentSpreedly's transaction status.
#----------------------------------------------------------------------------------------
# input       $Status              | New transaction status | String
#========================================================================================
sub updateTransactionStatus {
    my $self = shift;
    my ($Status, $hTransaction) = @_;

    $self->set( { 'TransStatus'  => $Status } );
    $self->set( { 'TransTimeMod' => $self->_getCurrentDateTime() } );
    $self->set( { 'TransID' => $self->get('GUID') } );
    unless ( $self->get('TransTime') ) {
        $self->set( { 'TransTime' => $self->get('TransTimeMod') } );
    }

    if ( isValid($hTransaction->{'transactionNo'}) ) {
        $self->set( { 'TransID2'  => $hTransaction->{'transactionNo'} } );
    }

    if ($hTransaction->{'paymentMethodType'} eq CREDIT_CARD){
        $self->set( { 'CardType'  => $hTransaction->{'card_type'} } );
    }

    return;
}

#========================================================================================
# function     _isPaymentSuccessful
# state        private
#----------------------------------------------------------------------------------------
# syntax       _isPaymentSuccessful($Status);
#----------------------------------------------------------------------------------------
# description  Returns a boolean regarding the status of the purchase.
#----------------------------------------------------------------------------------------
# input       $isPaymentSuccessful  | Purchase status | Boolean
#========================================================================================
sub isPaymentSuccessful {
    my $self = shift;

    my $TransStatus = $self->get('TransStatus');
    my $isPayed = (defined $TransStatus && ($TransStatus eq PAYMENT_STATUS_PAID));
    my $isTransIdSet = isValid($self->get('TransID'));
    my $isTransactionNumberSet = isValid($self->get('TransID2'));

    return ($isPayed && $isTransIdSet && $isTransactionNumberSet);
}

#========================================================================================
# function     _setOrderToPaid
# state        private
#----------------------------------------------------------------------------------------
# syntax       setOrderToPaid();
#----------------------------------------------------------------------------------------
# description  Sets Order to paid if the payment was successful.
#========================================================================================
sub setOrderToPaid {
    my $self = shift;

    return unless $self->isPaymentSuccessful();

    my $LineItemContainer = $self->container();
    my $Order = $LineItemContainer->parent();

    return unless $Order->instanceOf('CustomerOrder');

    $Order->set({'PaidOn' => $self->get('TransTime')});

    return;
}

#========================================================================================
# function     executePayment
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSpreedly->executePayment($hParams);
#----------------------------------------------------------------------------------------
# description  This function directs the payment process.
#----------------------------------------------------------------------------------------
# input        $hValues | environment information
#                       <ul>
#                           <li>Basket - basket object (optional)</li>
#                           <li>Servlet - servlet handling the current request</li>
#                           <li>Order - order object (optional)</li>
#                       </ul>
#                   | ref.hash
#========================================================================================
sub executePayment {
    my $self = shift;
    my ($hParams) = @_;

    my $Servlet = $hParams->{'Servlet'};

    # TODO: Maybe we need to do the same for both classical and mobile checkout.
    ($Servlet->vars('IsMobile')) ? $self->SUPER::updateOnExecute($hParams) : $self->SUPER::executePayment($hParams);

    return;
}

#========================================================================================
# function     getDisplayTransID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DisplayTransID = $LineItemPayment->getDisplayTransID() ;
#----------------------------------------------------------------------------------------
# description  This overwrites the default LineItemPaymentExternal->getDisplayTransID()
#               method to return TransID2 instead the default TransID.
#----------------------------------------------------------------------------------------
# return       $DisplayTransID | TransID2 attribute | string
#========================================================================================
sub getDisplayTransID {
    my $self = shift;

    return $self->get('TransID2');
}

#----------------------------------------------------------------------------------------
# Wrappers
#----------------------------------------------------------------------------------------
sub _getCurrentDateTime {
    my $self = shift;

    return GetCurrentDateTime();
}


#========================================================================================
# function     getTransactionToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $token = $LineItemPaymentSpreedly->getTransactionToken();
#----------------------------------------------------------------------------------------
# description  get the Transaction token from the line item payment object.
#----------------------------------------------------------------------------------------
# return       $TransactionToken | Transaction token string | string
#========================================================================================
sub getTransactionToken {
    my $self = shift;

    return $self->get('TransactionToken');
}

#========================================================================================
# function     setToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSpreedly->setTransactionToken( $TransactionToken );
#----------------------------------------------------------------------------------------
# description  save the Transaction token into the line item payment object.
#----------------------------------------------------------------------------------------
# input        $TransactionToken | Transaction token string | string
#========================================================================================
sub setTransactionToken {
    my $self = shift;
    my ($TransactionToken) = @_;

    $self->set({ 'TransactionToken' =>  $TransactionToken });

    return;
}

1;