#========================================================================================
# package      DE_EPAGES::Spreedly::UI::JSON::BasketOrder
# base         DE_EPAGES::Presentation::UI::JSON
# state        private
#----------------------------------------------------------------------------------------
# description  UI Controller for basket or order
#========================================================================================
package DE_EPAGES::Spreedly::UI::JSON;
use base DE_EPAGES::Presentation::UI::JSON;

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Order::API::Object::BasketOrderProcess;
use DE_EPAGES::Spreedly::API::PaymentInformationController;
use DE_EPAGES::Spreedly::API::PurchaseController;
use DE_EPAGES::Spreedly::API::Constants qw ( 
    PURCHASE_RESPONSE_STATE_SUCEEDED
    PURCHASE_RESPONSE_STATE_PENDING
    PURCHASE_RESPONSE_REQUIRED_ACTION_REDIRECT
    PAYMENT_STATUS_PAID
    JSON_FAILURE_MESSAGES
    THREE_DS_AVAILABLE_GATEWAYS );
use JSON qw ( decode_json );

use feature 'switch';

use strict;

#========================================================================================
# function     MakePurchase
# state        public
#----------------------------------------------------------------------------------------
# syntax       MakePurchase( $Servlet );
#----------------------------------------------------------------------------------------
# description  UI Controller for purchase at multi-step checkout
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub MakePurchase {
    my $self = shift;
    my ($Servlet) = @_;

    my $hReturnData = {
        'Error' => {
            'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
            'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'UnknownError'},
        }
    };

    my $Basket = $Servlet->object();
    my $LineItemContainer = $Basket->get('LineItemContainer');
    my $SpreedlyBrowserInfo = $Servlet->params()->{'SpreedlyBrowserInfo'};

    my $hTransaction;

    eval { $hTransaction = $self->_makePurchase( $LineItemContainer, $Basket, $SpreedlyBrowserInfo); };
    if ( ExistsError() ) {
        $hReturnData = $self->_HandleCriticalErrorOnRequest($Basket, $Servlet, 'Make Purchase Request die ocurred');
        return $self->_writeDataToOutput($hReturnData);
    }

    $self->_LogPayment($Servlet, 'Make Purchase', { 'Basket' =>  $Basket->guid(), 'Transaction' => $hTransaction });

    if ( $hTransaction->{'successful'} ) {
        $self->_SaveTransactionToken($hTransaction->{'token'}, $LineItemContainer);
        $hReturnData = $self->_HandleSuccessfulPurchaseResponse($hTransaction, $Basket, $Servlet);
    } else {
        $self->_LogPayment($Servlet, 'Purchase Response was not sucessful',$hTransaction->{'successful'});

        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'ConnectionError'},
            }
        };
    }
    return $self->_writeDataToOutput($hReturnData);
}

#========================================================================================
# function     CompleteTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       CompleteTransaction( $Servlet );
#----------------------------------------------------------------------------------------
# description  UI Controller for CompleteTransaction at multi-step checkout
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub CompleteTransaction {
    my $self = shift;
    my ($Servlet) = @_;

    my $BasketOrder = $Servlet->object();
    my $TransactionToken = $Servlet->params()->{'TransactionToken'};

    my $hTransaction;
    my $hResult;
    my $IsThreeDSAvailable = $self->_isThreeDSAvailable($BasketOrder);

    if (!$IsThreeDSAvailable) {
        $hResult = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Complete'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
        $self->_LogPayment($Servlet, 'Trying to call Complete Transaction on non 3DS Gateway', $hResult);
        return $self->_writeDataToOutput($hResult);
    }

    eval { $hTransaction = $self->_completeTransaction($BasketOrder, $TransactionToken); };
    if ( ExistsError() ) {
        $hResult = $self->_HandleCriticalErrorOnRequest($BasketOrder, $Servlet, 'Complete Transaction Error');
        return $self->_writeDataToOutput($hResult);
    }

    $self->_LogPayment($Servlet, 'Complete transaction', { 'Basket' =>  $BasketOrder->guid(), 'Transaction' => $hTransaction });

    if ( $hTransaction->{'successful'} ) {

        if ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED) {
            $self->_SaveTransactionNumber($hTransaction, $BasketOrder);
        }
        $hResult = {
            'Transaction' =>  {
                'state' => $hTransaction->{'state'},
                'required_action' => $hTransaction->{'required_action'},
                'challenge_url' => $hTransaction->{'challenge_url'},
                'device_fingerprint_form' => $hTransaction->{'device_fingerprint_form'},
            }
        };

    } else {
        $self->_LogPayment($Servlet, 'Complete Transaction was not sucessful',$hTransaction->{'successful'});
        $hResult = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Complete'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
    }

    return $self->_writeDataToOutput($hResult);
}

#----------------------------------------------------------------------------------------
sub _completeTransaction {
    my $self = shift;
    my ($BasketOrder, $TransactionToken) = @_;

    my $Shop        = $BasketOrder->getSite();
    my $GatewayName = $self->_getGatewayName($BasketOrder);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();
    my $hTransaction = $PurchaseController->CompleteTransaction($Shop, $GatewayName, $TransactionToken);

    return $hTransaction;
}
#----------------------------------------------------------------------------------------
sub _HandleCriticalErrorOnRequest {
    my $self = shift;
    my ($Basket, $Servlet, $LogMessage) = @_;

    my $Error = GetError();

    my $Response = $Error->vars->{Response};

    my $Reason = JSON_FAILURE_MESSAGES->{'Reason'}->{'UnknownError'};
    my $ErrorMessage = '';

    # Not an empty string and json format
    if(!$Response =~ /^$/ && $Response =~ /^\{/) {
        my $hResponse = decode_json($Response);

        $Reason = $hResponse->{'code'};
        $ErrorMessage = $hResponse->{'message'};
    }

    $self->_LogPayment($Servlet, $LogMessage,
        {
            'Basket' => $Basket->guid(),
            'ErrorMessageEp6' => $Error->message,
            'ErrorMessageEpJ' => $ErrorMessage,
        }
    );

    my $hReturnData = {
        'Error' => {
            'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
            'Reason' => $Reason
        }
    };

    return $hReturnData;
}

#----------------------------------------------------------------------------------------
sub _HandleSuccessfulPurchaseResponse {
    my $self = shift;
    my ($hTransaction, $Basket, $Servlet) = @_;

    my $hReturnData;

    if ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED) {
        $self->_SaveTransactionNumber( $hTransaction, $Basket );
        $hReturnData = {
            'Transaction' =>  {
                'state' => $hTransaction->{'state'},
                'required_action' => $hTransaction->{'required_action'},
            }
        };

    } elsif ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_PENDING && 
        $hTransaction->{'required_action'} eq PURCHASE_RESPONSE_REQUIRED_ACTION_REDIRECT) {

        $self->_LogPayment($Servlet, 'Purchase Response - We cannot handle 3DS1 flow',$hTransaction->{'required_action'});
        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };

    } elsif ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_PENDING && $self->_isThreeDSAvailable($Basket)) {
        $hReturnData = {
            'Transaction' =>  {
                'state' => $hTransaction->{'state'},
                'token' => $hTransaction->{'token'},
                'required_action' => $hTransaction->{'required_action'},
                'device_fingerprint_form' => $hTransaction->{'device_fingerprint_form'},
                'checkout_form' => $hTransaction->{'checkout_form'},
                'checkout_url' => $hTransaction->{'checkout_url'},
            }
        };

    } else {
        $self->_LogPayment($Servlet, 'Purchase Response - value of state unknown or not 3DS and pending',$hTransaction);
        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
    }

    return $hReturnData;
}

#========================================================================================
# function     _isThreeDSAvailable
# state        private
#----------------------------------------------------------------------------------------
# syntax       _isThreeDSAvailable();
#----------------------------------------------------------------------------------------
# description  Returns a boolean regarding the availability of 3DS on the gateway.
#----------------------------------------------------------------------------------------
# input       $IsThreeDSAvailable  | ThreeDS availability | Boolean
#========================================================================================
sub _isThreeDSAvailable {
    my $self = shift;
    my ( $Basket ) = @_;

    my $GatewayName = $self->_getGatewayName($Basket);
    if (!isValid($GatewayName)) {
        return 0;
    }
    if ($GatewayName ~~ THREE_DS_AVAILABLE_GATEWAYS) {
        return 1;
    }

    return 0;
}
#----------------------------------------------------------------------------------------
sub _SaveTransactionToken {
    my $self = shift;
    my ( $TransactionToken, $LineItemContainer ) = @_;

    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->setTransactionToken($TransactionToken);

    return;
}

#----------------------------------------------------------------------------------------
sub _SaveTransactionNumber {
    my $self = shift;
    my ( $hTransaction, $Basket ) = @_;

    my $LineItemContainer       = $Basket->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->updateTransactionStatus(PAYMENT_STATUS_PAID, $hTransaction);

    return;
}

#----------------------------------------------------------------------------------------
# Wrappers
#----------------------------------------------------------------------------------------
sub _getGatewayName {
    my $self = shift;
    my ( $Basket ) = @_;

    my $LineItemContainer = $Basket->get('LineItemContainer');
    return undef unless isValid($LineItemContainer);

    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    return undef unless isValid($LineItemPaymentSpreedly);

    my $PaymentMethod = $LineItemPaymentSpreedly->get('PaymentMethod');
    return undef unless isValid($PaymentMethod);

    my $PaymentType = $PaymentMethod->get('PaymentType');
    return undef unless isValid($PaymentType);

    return $PaymentType->get('GatewayName');
}

#----------------------------------------------------------------------------------------
sub _LogPayment {
    my $self = shift;
    my ($Servlet, $Name, $DebugStuff) = @_;

    my $Basket = $Servlet->object();
    my $LineItemPayment = $self->_RetrieveLineItemPayment($Basket);
    my $PaymentMethod   = $LineItemPayment->get('PaymentMethod');
    my $PaymentTypeAlias = $PaymentMethod->get('PaymentType')->alias();

    LogPayment($PaymentTypeAlias, $Name, $DebugStuff);

    return;
}

#----------------------------------------------------------------------------------------
sub _makePurchase {
    my $self = shift;
    my ( $LineItemContainer, $Basket, $SpreedlyBrowserInfo) = @_;

    my $Shop             = $Basket->getSite();
    my $GatewayName = $self->_getGatewayName($Basket);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();

    return $PurchaseController->makePurchase( $LineItemContainer, $Shop, $GatewayName, $SpreedlyBrowserInfo);
}

#----------------------------------------------------------------------------------------
sub _RetrieveLineItemPayment {
    my $self = shift;
    my ($Basket) = @_;

    my $LineItemContainer = $Basket->container();

    return $LineItemContainer->get('Payment');
}
#----------------------------------------------------------------------------------------
sub _writeDataToOutput {
    my $self = shift;
    my ($hData) = @_;

    $self->writeContentOutput($hData);
    return;
}

1;
